@@ -1,5 +1,5 @@
 #
-#  Copyright (c) 2019-2022, ETH Zurich. All rights reserved.
+#  Copyright (c) 2019-2023, ETH Zurich. All rights reserved.
 #
 #  Please, refer to the LICENSE file in the root directory.
 #  SPDX-License-Identifier: BSD-3-Clause
@@ -91,6 +91,14 @@ class UnexpectedStatusException(FirecrestException):
         return f"{super().__str__()}: expected status {self._expected_status_code}"
 
 
+class NoJSONException(FirecrestException):
+    """Exception raised when JSON in not included in the response
+    """
+
+    def __str__(self):
+        return f"{super().__str__()}: JSON is not included in the response"
+
+
 class StorageDownloadException(FirecrestException):
     """Exception raised by a failed external download
     """
