@@ -1,5 +1,5 @@
 #
-#  Copyright (c) 2019-2022, ETH Zurich. All rights reserved.
+#  Copyright (c) 2019-2023, ETH Zurich. All rights reserved.
 #
 #  Please, refer to the LICENSE file in the root directory.
 #  SPDX-License-Identifier: BSD-3-Clause
@@ -62,15 +62,22 @@ custom_theme = {
     "repr.number": "none",
 }
 console = Console(theme=Theme(custom_theme))
-client = None
+client: fc.Firecrest = None  # type: ignore
+logger = logging.getLogger(__name__)
 
 
-def examine_exeption(e):
-    msg = "{__app_name__}: Operation failed"
+def examine_exeption(e: Exception) -> None:
+    msg = f"{__app_name__}: Operation failed"
     if isinstance(e, fc.ClientsCredentialsException):
         msg += ": could not fetch token"
+    elif isinstance(e, fc.FirecrestException):
+        msg += ": a Firecrest client error has occurred"
+    else:
+        # in case of FirecrestException and ClientsCredentialsException
+        # we don't need to log again the exception
+        logger.critical(e)
 
-    console.print(f"[red]{__app_name__}: Operation failed[/red]")
+    console.print(f"[red]{msg}[/red]")
 
 
 def create_table(table_title, data, *mappings):
@@ -114,7 +121,7 @@ def services(
             ("Description", "description"),
         )
         console.print(table, overflow="fold")
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -143,7 +150,7 @@ def systems(
             ("Description", "description"),
         )
         console.print(table, overflow="fold")
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -171,7 +178,7 @@ def parameters():
         )
 
         console.print(storage_table, utilities_table, overflow="fold")
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -208,7 +215,7 @@ def tasks(
         else:
             console.print(table)
 
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -248,7 +255,7 @@ def ls(
             )
 
             console.print(table)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -269,7 +276,7 @@ def mkdir(
     """
     try:
         client.mkdir(machine, path, p)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -286,7 +293,7 @@ def mv(
     """
     try:
         client.mv(machine, source, destination)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -303,7 +310,7 @@ def chmod(
     """
     try:
         client.chmod(machine, path, mode)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -323,7 +330,7 @@ def chown(
     """
     try:
         client.chown(machine, path, owner, group)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -340,7 +347,7 @@ def cp(
     """
     try:
         client.copy(machine, source, destination)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -356,7 +363,7 @@ def file(
     """
     try:
         console.print(client.file_type(machine, path))
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -442,7 +449,7 @@ def stat(
             )
 
             console.print(table)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -459,7 +466,7 @@ def symlink(
     """
     try:
         client.symlink(machine, target, link_name)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -482,7 +489,7 @@ def rm(
             client.simple_delete(machine, path)
         else:
             console.print("Operation cancelled")
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -498,7 +505,7 @@ def checksum(
     """
     try:
         console.print(client.checksum(machine, path))
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -509,15 +516,62 @@ def head(
         ..., help="The machine name where the filesystem belongs to."
     ),
     path: str = typer.Argument(..., help="The absolute target path."),
+    lines: int = typer.Option(
+        None, "-n", "--lines", help="Print count lines of each of the specified files."
+    ),
+    bytes: int = typer.Option(
+        None, "-c", "--bytes", help="Print bytes of each of the specified files."
+    ),
 ):
-    """View the content of a specified file
+    """Display the beginning of a specified file.
+    By default the first 10 lines will be returned.
+    Bytes and lines cannot be specified simultaneously.
 
     You view only files smaller than UTILITIES_MAX_FILE_SIZE bytes.
     This variable is available in the parameters command.
     """
+    if lines and bytes:
+        console.print(
+            f"[red]{__app_name__} head: cannot specify both 'bytes' and 'lines'[/red]"
+        )
+        raise typer.Exit(code=1)
+
     try:
-        console.print(client.view(machine, path))
-    except fc.FirecrestException as e:
+        console.print(client.head(machine, path, bytes, lines))
+    except Exception as e:
+        examine_exeption(e)
+        raise typer.Exit(code=1)
+
+
+@app.command(rich_help_panel="Utilities commands")
+def tail(
+    machine: str = typer.Argument(
+        ..., help="The machine name where the filesystem belongs to."
+    ),
+    path: str = typer.Argument(..., help="The absolute target path."),
+    lines: int = typer.Option(
+        None, "-n", "--lines", help="Print count lines of each of the specified files."
+    ),
+    bytes: int = typer.Option(
+        None, "-c", "--bytes", help="Print bytes of each of the specified files."
+    ),
+):
+    """Display the end of a specified file.
+    By default the last 10 lines will be returned.
+    Bytes and lines cannot be specified simultaneously.
+
+    You view only files smaller than UTILITIES_MAX_FILE_SIZE bytes.
+    This variable is available in the parameters command.
+    """
+    if lines and bytes:
+        console.print(
+            f"[red]{__app_name__} tail: cannot specify both 'bytes' and 'lines'[/red]"
+        )
+        raise typer.Exit(code=1)
+
+    try:
+        console.print(client.tail(machine, path, bytes, lines))
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -528,7 +582,7 @@ def whoami():
     """
     try:
         console.print(client.whoami())
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -552,14 +606,16 @@ def download(
         TransferType.direct,
         "--type",
         case_sensitive=False,
-        help=f"Select type of transfer (run `{__app_name__} download --help` for details).",
+        help=f"Select type of transfer.",
     ),
 ):
     """Download a file
 
-    Direct download will download the file to the DESTINATION but it will work only for small files. You can find the maximum size by running the `parameters` command.
+    Direct download will download the file to the DESTINATION but it will work only for small files.
+    You can find the maximum size in UTILITIES_MAX_FILE_SIZE by running the `parameters` command.
 
-    External download will return with a link in case of success. The file can be downloaded locally from there without any authentication.
+    External download will return with a link in case of success.
+    The file can be downloaded locally from there without any authentication.
     """
     try:
         if transfer_type == TransferType.direct:
@@ -578,7 +634,7 @@ def download(
                 "cancel the command and follow up through the task."
             ):
                 console.out(f"Download the file from:\n{down_obj.object_storage_data}")
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -594,16 +650,22 @@ def upload(
     ),
     filename: Optional[str] = typer.Argument(
         None,
-        help="The name of the file in the machine (by default it will be same as the local file).",
+        help="The name of the file in the machine (by default it will be same as the local file). It works only for a direct upload.",
     ),
     transfer_type: TransferType = typer.Option(
         TransferType.direct,
         "--type",
         case_sensitive=False,
-        help=f"Select type of transfer (run `{__app_name__} upload --help` for details).",
+        help=f"Select type of transfer.",
     ),
 ):
     """Upload a file
+
+    Direct upload will upload the file to the DESTINATION directory but it will work only for small files.
+    You can find the maximum size in UTILITIES_MAX_FILE_SIZE by running the `parameters` command.
+
+    External download will return with a command that will need to be run in case of success.
+    The file can be uploaded to a stage area without any authentication and FirecREST will move the file to the cluster's filesystem as soon as it finished.
     """
     try:
         if transfer_type == TransferType.direct:
@@ -621,16 +683,16 @@ def upload(
                 data = up_obj.object_storage_data
                 if "command" in data:
                     console.print(
-                        "\nRun the following the following command to finish the upload:"
-                    )
-                    console.out(data["command"])
-                    console.print(
-                        "\nYou can also use a different software to upload the file:"
+                        "\nNecessary information to upload the file in the staging area:"
                     )
                     console.print(f"[yellow]{data['parameters']}[/yellow]")
+                    console.print(
+                        "\nOr simply run the following command to finish the upload:"
+                    )
+                    console.print(f"[green]{data['command']}[/green]")
                 else:
                     console.print(data)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -644,6 +706,11 @@ def submit(
         ...,
         help="The path of the script (if it's local it can be relative path, if it is on the machine it has to be the absolute path)",
     ),
+    account: Optional[str] = typer.Option(
+        None,
+        "--account",
+        help="Charge resources used by this job to specified account.",
+    ),
     local: Optional[bool] = typer.Option(
         True,
         help="The batch file can be local (default) or on the machine's filesystem.",
@@ -652,8 +719,8 @@ def submit(
     """Submit a batch script to the workload manger of the target system
     """
     try:
-        console.print(client.submit(machine, job_script, local))
-    except fc.FirecrestException as e:
+        console.print(client.submit(machine, job_script, local, account=account))
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -693,7 +760,7 @@ def submit_mv(
                 machine, source, destination, job_name, time, jobid, account
             )
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -733,7 +800,7 @@ def submit_cp(
                 machine, source, destination, job_name, time, jobid, account
             )
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -773,7 +840,7 @@ def submit_rsync(
                 machine, source, destination, job_name, time, jobid, account
             )
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -810,7 +877,7 @@ def submit_rm(
         console.print(
             client.submit_delete_job(machine, path, job_name, time, jobid, account)
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -856,7 +923,7 @@ def poll(
                 ("User", "user"),
             )
             console.print(table)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -894,7 +961,7 @@ def poll_active(
                 ("User", "user"),
             )
             console.print(table)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -910,7 +977,7 @@ def cancel(
     """
     try:
         client.cancel(machine, job)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -926,7 +993,7 @@ def list(
     try:
         res = client.all_reservations(machine)
         console.print(res)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -957,7 +1024,7 @@ def create(
         client.create_reservation(
             machine, name, account, num_nodes, node_type, start_time, end_time
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -988,7 +1055,7 @@ def update(
         client.update_reservation(
             machine, name, account, num_nodes, node_type, start_time, end_time
         )
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -1004,7 +1071,7 @@ def delete(
     """
     try:
         client.delete_reservation(machine, name)
-    except fc.FirecrestException as e:
+    except Exception as e:
         examine_exeption(e)
         raise typer.Exit(code=1)
 
@@ -1035,6 +1102,14 @@ def main(
     verbose: Optional[bool] = typer.Option(
         None, "-v", "--verbose", help="Enable verbose mode."
     ),
+    timeout: Optional[float] = typer.Option(
+        None,
+        help="How many seconds to wait for the FirecREST server to send data before giving up.",
+    ),
+    auth_timeout: Optional[float] = typer.Option(
+        None,
+        help="How many seconds to wait for the authorization server to send data before giving up.",
+    ),
     debug: Optional[bool] = typer.Option(None, help="Enable debug mode."),
 ):
     """
@@ -1048,7 +1123,9 @@ def main(
     """
     global client
     auth_obj = fc.ClientCredentialsAuth(client_id, client_secret, token_url)
+    auth_obj.timeout = auth_timeout
     client = fc.Firecrest(firecrest_url=firecrest_url, authorization=auth_obj)
+    client.timeout = timeout
     if debug:
         logging.basicConfig(
             level=logging.DEBUG,
